import pwd
from typing import Any, Dict, TypedDict

from lxcraft import Plan
from lxcraft.user import User

USERNAME = "lxcraft"


class UserFields(TypedDict):
    username: str
    gecos: str | None
    password: str | None
    shell: str
    home: str | None
    uid: int | None
    gid: int | None
    groups: list | None


fields: UserFields = {
    "username": USERNAME,
    "gecos": "lxcraft",
    "password": "lxcraft",
    "shell": "/bin/sh",
    "home": "/home/lxcraft_test",
    "uid": 2000,
    "gid": 100,
    "groups": ["sudo"],
}


def test_user_add():
    kwargs: Dict[str, Any] = {}
    for key, value in fields.items():
        kwargs[key] = value
        Plan([User(**kwargs)]).execute()
        assert pwd.getpwnam(USERNAME)

    # Idempotency test
    Plan([User(**kwargs)]).execute()
    assert pwd.getpwnam(USERNAME)

    Plan([User(**kwargs)]).destroy()



from pathlib import Path

from lxcraft import Plan
from lxcraft.path import Directory

DIRNAME = "/run/user/www-data"


def test_directory_create():
    with Plan([Directory(DIRNAME)]) as plan:
        plan.execute()
        assert Path(DIRNAME).is_dir()

        Plan([Directory(DIRNAME, owner_user="bin")]).execute()
        Plan([Directory(DIRNAME, owner_group="bin")]).execute()

        Plan([Directory(DIRNAME, "bin", "bin", 0o755)]).execute()



import pytest

from lxcraft import Plan
from lxcraft.debian import AptPackages


def test_apt_package_ok():
    with Plan([AptPackages(["nginx"])]) as plan:
        plan.execute()
        Plan([AptPackages(["nginx"])]).execute()  # test idempotency


def test_apt_package_fail():
    with pytest.raises(Exception, match=r"Command terminated with non zero exit code"):
        Plan([AptPackages(["blaldderandom"])]).execute()



#!/usr/bin/env bash
import os

base_dir = os.environ.get("USERPROFILE", os.getcwd())

cmd = (
    "docker run"
    f" -v {base_dir}:{base_dir}"
    " -v $HOME/.gitconfig:/root/.gitconfig -v $HOME/.ssh:/root/.ssh"
    " -it debian:12 $(pwd)/develop/setup.sh $(pwd)"
)
print(cmd)
os.system(cmd)



from dataclasses import dataclass
from subprocess import getstatusoutput

import lxcraft


@dataclass
class PipPackages(lxcraft.Resource):
    """List of packages that must be installed"""

    package_list: list[str]

    def create(self):
        package_list = " ".join(self.package_list)
        lxcraft.system(f"pip install {package_list}")

    def destroy(self):
        package_list = " ".join(self.package_list)
        lxcraft.system(f"pip uninstall -y {package_list}")

    def is_created(self):
        for package in self.package_list:
            if not self.is_installed(package):
                return False
        return True

    def is_consistent(self):
        # TODO: Add support for version checking
        return True

    @staticmethod
    def is_installed(package_name: str):
        rc, _ = getstatusoutput(f"pip show {package_name}")
        return rc == 0



import os

from taglogger import tlog


def system(*args, ignore_errors=False, **kwargs):
    """Run a shell command"""
    rc = os.system(*args, **kwargs)
    tlog(
        "command",
        *args,
        **kwargs,
    )
    tlog("command", f"# rc={rc}, ignore_errors={ignore_errors}")
    if not ignore_errors and rc != 0:
        raise Exception(f"Command terminated with non zero exit code {rc}")


# def get_output(*args, **kwargs):
#     """Run a shell command"""
#     rc, output = subprocess.getstatusoutput(*args, **kwargs)
#     tlog(
#         "command",
#         *args,
#         f"# rc= {rc}",
#         **kwargs,
#     )
#     if rc != 0:
#         raise Exception(f"Command terminated with non zero exit code {rc}")
#     return output



import sys
from dataclasses import dataclass, field
from typing import Callable

from taglogger import tlog

from .resource import Resource


@dataclass
class Plan:
    resources: list[Resource] = field(
        default_factory=list[Resource]
    )  # List of plan resources

    # Check if all the resources are based on Resource to avoid type errors
    def __post_init__(self):
        assert self.resources, "Plan must have at least one resource"
        # for resource in self.resources:
        #     assert isinstance(
        #         resource, Resource
        #     ), f"{resource} is not based on Resource"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for resource in self.resources:
            tlog("destroy", "__exit__ Destroying", resource)
            resource.destroy()  # type: ignore[attr-defined]

    def try_and_point(self, action: Callable):
        tlog("action", "Trying to execute", action)
        try:
            action()
        except Exception as e:
            print(
                "EXCEPTION related to ",
                action.__self__.source_repr(),  # type: ignore[attr-defined]
                file=sys.stderr,
            )
            raise e

    def execute(self):
        """Execute all the actions required to create the resources
        and bring them to a consistent state"""
        for resource in self.resources:
            if not resource.is_created():  # type: ignore[attr-defined]
                tlog("action", "Creating missing", resource)
                self.try_and_point(resource.create)  # type: ignore[attr-defined]
                on_change_callback = getattr(resource, "on_change_callback", None)
                if on_change_callback:
                    tlog("action", "on_change", resource, on_change_callback)
                    on_change_callback()
            if not resource.is_consistent():  # type: ignore[attr-defined]
                self.try_and_point(resource.destroy)  # type: ignore[attr-defined]
                self.try_and_point(resource.create)  # type: ignore[attr-defined]

    def destroy(self):
        """Destroy all the resources"""
        for resource in self.resources:
            tlog("destroy", "Destroying all resources")
            for resource in self.resources:
                if resource.is_created():  # type: ignore[attr-defined]
                    self.try_and_point(resource.destroy)  # type: ignore[attr-defined]



import inspect
from typing import Callable


class Resource:
    def __post_init__(self):
        callerframerecord = inspect.stack()[2]
        self.location = callerframerecord.filename
        self.lineno = callerframerecord.lineno

    def source_repr(self):
        return f"{self.location}:{self.lineno} # -> {self}"

    def on_change(self, callback: Callable):
        """Execute callback if the resource is changed"""
        self.on_change_callback = callback
        return self



import os

from invoke import task


@task
def docker(c):
    c.run("sh develop/docker-bash", pty=True)


@task
def setup(c):
    print("Installing Python requirements...")
    c.run("pip install -r requirements.txt -r requirements-dev.txt", hide=True)


@task
def test(c):
    c.run("python -m pytest -x tests/", pty=True)


@task
def test_only(c, filter: str):
    c.run(f"python -m pytest -x -v -s -k {filter}", pty=True)


@task
def cover(c):
    c.run("python -m pytest -x --cov-report term-missing --cov=lxcraft tests", pty=True)


@task
def cover_only(c, filter: str):
    c.run(
        f"python -m pytest -x --cov-report term-missing --cov=lxcraft tests -k {filter}",
        pty=True,
    )


@task
def run_sample(c, filename: str):
    os.chdir("samples")
    os.environ["PYTHONPATH"] = ".."
    c.run(
        f"python ../{filename}",
        pty=True,
    )



from dataclasses import dataclass

import lxcraft


@dataclass
class MyTestresourceWithAction(lxcraft.Resource):
    pending_action: bool = True

    def create(self):
        self.pending_action = False

    def destroy(self):
        pass

    def is_created(self):
        return not self.pending_action

    def is_consistent(self):
        return True


def test_resource():
    def change_callback():
        pass

    resource = MyTestresourceWithAction().on_change(change_callback)
    assert resource.source_repr()
    with lxcraft.Plan([resource]) as plan:
        plan.execute()
        plan.execute()



import grp
import os
import pwd
from dataclasses import dataclass
from pathlib import Path

import lxcraft


@dataclass
class Directory(lxcraft.Resource):
    """Directory in the filesystem"""

    path: str
    owner_user: str = ""
    owner_group: str = ""
    mode: int = 0o755

    def create(self):
        Path(self.path).mkdir(parents=True, exist_ok=True)
        if self.owner_user != "" or self.owner_group != "":
            self.chown()
        if self.mode:
            self.chmod()

    def destroy(self):
        Path(self.path).rmdir()

    def is_created(self):
        return Path(self.path).exists()

    def is_consistent(self):
        return Path(self.path).is_dir() and self.mode_matches() and self.owner_matches()

    def owner_matches(self):
        current_uid = Path(self.path).stat().st_uid
        current_gid = Path(self.path).stat().st_gid
        if self.owner_user and pwd.getpwnam(self.owner_user).pw_uid != current_uid:
            return False
        if self.owner_group and grp.getgrnam(self.owner_group).gr_gid != current_gid:
            return False
        return True

    def mode_matches(self):
        return self.mode == Path(self.path).stat().st_mode & 0o777

    def chown(self):
        if self.owner_user == "":
            uid = os.getuid()
        else:
            uid = pwd.getpwnam(self.owner_user).pw_uid
        if self.owner_group == "":
            gid = os.getgid()
        else:
            gid = grp.getgrnam(self.owner_group).gr_gid
        os.chown(self.path, uid, gid)

    def chmod(self):
        os.chmod(self.path, self.mode)



from dataclasses import dataclass

import pytest

import lxcraft


class MyTestresource(lxcraft.Resource):
    def create(self):
        pass

    def destroy(self):
        pass

    def is_created(self):
        pass

    def is_consistent(self):
        pass


@dataclass
class MyTestresourceWithActions(lxcraft.Resource):
    raise_exeception: bool = False
    was_created: bool = False

    def create(self):
        if self.raise_exeception:
            raise Exception("TestresourceWithActionsException")
        self.was_created = True

    def destroy(self):
        pass

    def is_created(self):
        return self.was_created

    def is_consistent(self):
        return True


def test_plan():
    # with pytest.raises(Exception, match=r"is not based on Resource"):
    #     lxcraft.Plan("text")

    # Test the context manager
    with lxcraft.Plan([MyTestresource()]) as plan:
        plan.execute()
        plan.destroy()

    with lxcraft.Plan([MyTestresourceWithActions()]) as plan:
        plan.execute()
        plan.destroy()

    lxcraft.Plan([MyTestresourceWithActions()]).execute()

    with pytest.raises(Exception, match=r"TestresourceWithActionsException"):
        lxcraft.Plan([MyTestresourceWithActions(raise_exeception=True)]).execute()



import pytest

from lxcraft import Plan
from lxcraft.python import PipPackages


def test_pip_package():
    TEST_PACKAGE = "docopt"

    with Plan([PipPackages([TEST_PACKAGE])]) as plan:
        plan.execute()
        Plan([PipPackages([TEST_PACKAGE])]).execute()  # test idempotency
        assert PipPackages.is_installed(TEST_PACKAGE)

    with pytest.raises(Exception, match=r"Command terminated with non zero exit code"):
        Plan([PipPackages([".blaldderandom"])]).execute()



from pathlib import Path

import pytest

from lxcraft import Plan
from lxcraft.path import FileContent

FileContent.set_templates_directory("tests/templates")
SOURCE_PATH = "tests/templates/test.txt"
TARGET_PATH = "/tmp/test.txt"


def test_filecontent_create():
    with Plan([FileContent(TARGET_PATH)]) as plan:
        plan.execute()
        target_text = Path(TARGET_PATH).read_text()
        assert "this is a {word} test\n" == target_text

    with Plan([FileContent(TARGET_PATH, SOURCE_PATH)]) as plan:
        plan.execute()
        target_text = Path(TARGET_PATH).read_text()
        assert "this is a {word} test\n" == target_text

    with pytest.raises(FileNotFoundError):
        with Plan([FileContent(TARGET_PATH, "missing_file")]) as plan:
            plan.execute()


def test_filecontent_create_replace():
    with Plan([FileContent(TARGET_PATH, replace={"{word}": "bananas"})]) as plan:
        plan.execute()
        target_text = Path(TARGET_PATH).read_text()
        assert "bananas" in target_text


def test_filecontent_chmod():
    with Plan([FileContent(TARGET_PATH, mode=0o600)]) as plan:
        plan.execute()
        assert oct(Path(TARGET_PATH).stat().st_mode & 0o777) == "0o600"


def test_filencontnet_chown():
    # Adjust owner on creation
    with Plan([FileContent(TARGET_PATH, owner_user="daemon")]) as plan:
        plan.execute()
        assert Path(TARGET_PATH).stat().st_uid == 1

    # Adjust owner after creation
    Plan([FileContent(TARGET_PATH)]).execute()
    with Plan([FileContent(TARGET_PATH, owner_user="daemon")]) as plan:
        plan.execute()
        assert Path(TARGET_PATH).stat().st_uid == 1

    Plan([FileContent(TARGET_PATH)]).execute()
    with Plan([FileContent(TARGET_PATH, owner_group="bin")]) as plan:
        plan.execute()
        assert Path(TARGET_PATH).stat().st_gid == 2



import grp
import pwd
from dataclasses import dataclass, field

import lxcraft


@dataclass
class User(lxcraft.Resource):
    username: str
    gecos: str | None = None
    password: str | None = None
    shell: str = "/usr/sbin/nologin"
    home: str | None = None
    uid: int | None = None
    gid: int | None = None
    groups: list | None = field(default_factory=list)
    create_home: bool = True

    def create(self):
        self.create_user()

    def destroy(self):
        lxcraft.system(f"userdel {self.username}")

    def is_created(self):
        try:
            pwd.getpwnam(self.username)
        except KeyError:
            return False
        return True

    def is_consistent(self):
        user_data = pwd.getpwnam(self.username)
        if self.gecos is not None and self.gecos != user_data.pw_gecos:
            return False
        if self.shell is not None and self.shell != user_data.pw_shell:
            return False
        if self.home is not None and self.home != user_data.pw_dir:
            return False
        if self.uid is not None and self.uid != user_data.pw_uid:
            return False
        if self.gid is not None and self.gid != user_data.pw_gid:
            return False
        if self.groups is not None and self.groups != self.current_user_groups():
            return False
        return True

    def current_user_groups(self):
        return [g.gr_name for g in grp.getgrall() if self.username in g.gr_mem]

    def create_user(self):
        cmd = f"useradd {self.username}"
        if self.gecos is not None:
            cmd += f" -c {self.gecos}"
        if self.password is not None:
            cmd += f" -p {self.password}"
        if self.shell is not None:
            cmd += f" -s {self.shell}"
        if self.home is not None:
            cmd += f" -d {self.home}"
        if self.uid is not None:
            cmd += f" -u {self.uid}"
        if self.gid is not None:
            cmd += f" -g {self.gid}"
        if self.groups:
            cmd += f" -G {','.join(self.groups)}"
        if self.create_home:
            cmd += " -m"
        lxcraft.system(cmd)



import grp
import os
import pwd
from dataclasses import dataclass, field
from pathlib import Path

import lxcraft


@dataclass
class FileContent(lxcraft.Resource):
    """Create file with content"""

    target_path: str
    source_path: str = ""
    owner_user: str = ""
    owner_group: str = ""
    mode: int = 0o644
    replace: dict = field(default_factory=dict)
    templates_directory = "templates"

    @staticmethod
    def set_templates_directory(directory: str):
        FileContent.templates_directory = directory

    def create(self):
        Path(self.target_path).write_text(self.get_template_text())
        if self.owner_user != "" or self.owner_group != "":
            self.chown()
        if self.mode:
            self.chmod()

    def destroy(self):
        Path(self.target_path).unlink()

    def is_created(self):
        return Path(self.target_path).exists()

    def is_consistent(self):
        return (
            self.get_template_text() == Path(self.target_path).read_text()
            and self.mode_matches()
            and self.owner_matches()
        )

    def get_template_text(self):
        if self.source_path:
            source_path = Path(self.source_path)
        else:
            filename = Path(self.target_path).name
            source_path = Path(self.templates_directory) / filename
        if source_path.exists():
            source_text = source_path.read_text()
        else:
            raise FileNotFoundError(f"Source file {source_path} not found")

        replaced_text = source_text
        for key, value in self.replace.items():
            replaced_text = replaced_text.replace(key, value)
        return replaced_text

    def owner_matches(self):
        current_uid = Path(self.target_path).stat().st_uid
        current_gid = Path(self.target_path).stat().st_gid
        if self.owner_user and pwd.getpwnam(self.owner_user).pw_uid != current_uid:
            return False
        if self.owner_group and grp.getgrnam(self.owner_group).gr_gid != current_gid:
            return False
        return True

    def mode_matches(self):
        return self.mode == Path(self.target_path).stat().st_mode & 0o777

    def chown(self):
        if self.owner_user == "":
            uid = os.getuid()
        else:
            uid = pwd.getpwnam(self.owner_user).pw_uid
        if self.owner_group == "":
            gid = os.getgid()
        else:
            gid = grp.getgrnam(self.owner_group).gr_gid
        os.chown(self.target_path, uid, gid)

    def chmod(self):
        os.chmod(self.target_path, self.mode)
